###################
# whois - guess IP country location
#
# Copyright Jean-Michel Philippe (c) 2007
#   <http://philipjm.free.fr/blog/>
#
# This is open source software released under the GPL license.
# The full text of this license is found in the file 'LICENSE',
# included with this source code package.
###################

"""
whois Python module
-------------------

Guess host location from their IP, based on whois services.

2 method are implemented:

* IP searched in a local database
* IP searched on an Internet whois service

Search results are stored in a local database to speedup further identical
searches. Use of former results can be disabled. Internet whois search can
be disabled too.

Local databases can be downloaded/updated from:
   http://software77.net/cgi-bin/ip-country/geo-ip.pl
   http://ip-to-country.webhosting.info/node/view/6

Full Internet whois queries may also be accessed.

Usage
-----

Import whois module:

 >>> import whois
 > loaded 79552 IP ranges from /usr/lib/.../whois/db/IpToCountry.csv.db

Query host country:

 >>> print whois.guessIPcountry('194.109.137.218')
 NETHERLANDS

Full Internet whois query:

 >>> Whois = whois.WhoisConsumer('194.109.137.218')
 >>> whois.WhoisRequest(Whois, whois.WhoisServer)
 <whois._whois.WhoisRequest at -0x4a62ce74>
 >>> whois.asyncore.loop()
 >>> print Whois.text
 
 OrgName:    RIPE Network Coordination Centre
 OrgID:      RIPE
 Address:    P.O. Box 10096
 City:       Amsterdam
 StateProv:
 PostalCode: 1001EB
 Country:    NL
 
 ReferralServer: whois://whois.ripe.net:43
 
 NetRange:   194.0.0.0 - 194.255.255.255
 CIDR:       194.0.0.0/8
 NetName:    RIPE-CBLK2
 NetHandle:  NET-194-0-0-0-1
 Parent:
 NetType:    Allocated to RIPE NCC
 NameServer: NS-PRI.RIPE.NET
 NameServer: NS3.NIC.FR
 NameServer: SUNIC.SUNET.SE
 NameServer: NS-EXT.ISC.ORG
 NameServer: SEC1.APNIC.NET
 NameServer: SEC3.APNIC.NET
 NameServer: TINNIE.ARIN.NET
 Comment:    These addresses have been further assigned to users in
 Comment:    the RIPE NCC region. Contact information can be found in
 Comment:    the RIPE database at http://www.ripe.net/whois
 RegDate:    1993-07-21
 Updated:    2005-08-03
 
 # ARIN WHOIS database, last updated 2007-03-24 19:10
 # Enter ? for additional hints on searching ARIN's WHOIS database.

"""

###################
# importations
import os, csv, pylab, asyncore, re, cPickle, md5
from requester import WhoisRequest, WhoisConsumer
from types import *

###################
# constants
WhoisServer = "whois.arin.net"

###################
# versionning
__author__ = "jean-michel.philippe@libertysurf.fr"
__url__ = "http://philipjm.free.fr/blog/"
__version__ = "0.1.0"

###################
# inits
__ModulePath__ = os.path.dirname(__file__)
__DBfile__ = os.path.join(__ModulePath__, "db/searched-IPs.db")
#__IPCountryFile__ = os.path.join(__ModulePath__, "db/ip-to-country.csv")
__IPCountryFile__ = os.path.join(__ModulePath__, "db/IpToCountry.csv")
__DB__ = {}
IPCountryTable = {}
IPCountryList = []
FromIPList = []
ToIPList = []

###################
def readIPCountryFile(verbose=True):
	"""
	Reads an IP/country database file.
	
	>>> readIPCountryFile(verbose=True)
	
	Input:
		- verbose	= display some information
	"""
	global __IPCountryFile__, IPCountryTable, IPCountryList, FromIPList, ToIPList
	
	# compute the IP/country file MD5
	CSVFileMD5 = md5.new()
	FileId = open(__IPCountryFile__, "r")
	CSVFileMD5.update(FileId.read())
	FileId.close()
	
	# check for an existing converted file
	if os.path.isfile(__IPCountryFile__ + '.db'):
		FileId = open(__IPCountryFile__ + '.db', 'rb')
		DBFileMD5Digest = cPickle.load(FileId)
		if CSVFileMD5.digest() == DBFileMD5Digest:
			IPCountryTable = cPickle.load(FileId)
			IPCountryList = cPickle.load(FileId)
			FromIPList = cPickle.load(FileId)
			ToIPList = cPickle.load(FileId)
		FileId.close()
		if CSVFileMD5.digest() == DBFileMD5Digest:
			print "> loaded %d IP ranges from %s" % (len(IPCountryList), __IPCountryFile__ + '.db')
			return
	
	# read from CSV
	FileId = open(__IPCountryFile__, "rb")
	CSVReader = csv.reader(FileId)
	for Row in CSVReader:
		if not len(Row) or Row[0][0] == '#':
			continue
		FromIP, ToIP = Row[:2]
		CountryCode, dummy, CountryName = Row[-3:]
		IPCountryTable[CountryCode] = CountryName
		IPCountryList.append( CountryName )
		FromIPList.append( int(FromIP) )
		ToIPList.append( int(ToIP) )
	FileId.close()
	FromIPList = pylab.asarray(FromIPList)
	ToIPList = pylab.asarray(ToIPList)
	if verbose:
		print "> loaded %d IP ranges from %s" % (len(IPCountryList), __IPCountryFile__)
	
	# save converted file
	FileId = open(__IPCountryFile__ + '.db', 'wb')
	Data = cPickle.dump(CSVFileMD5.digest(), FileId)
	Data = cPickle.dump(IPCountryTable, FileId)
	Data = cPickle.dump(IPCountryList, FileId)
	Data = cPickle.dump(FromIPList, FileId)
	Data = cPickle.dump(ToIPList, FileId)
	FileId.close()

###################
def loadDatabase():
	"""
	Loads the database of previously searched IP.
	
	>>> loadDatabase()
	"""
	global __DBfile__, __DB__
	
	if not os.path.isfile(__DBfile__):
		return False
	FileId = open(__DBfile__, 'rb')
	__DB__ = cPickle.load(FileId)
	FileId.close()
	return True

###################
def saveDatabase():
	"""
	Saves the searched IP database.
	
	>>> Data = saveDatabase()
	
	Output:
		- Data	= dictionary Data[IP] = country
	
	NB: IP is an integer number as returned by IP2num
	"""
	global __DBfile__, __DB__
	
	FileId = open(__DBfile__, 'wb')
	Data = cPickle.dump(__DB__, FileId)
	FileId.close()
	return Data

###################
def addDatabase(IP, CountryName):
	"""
	Records a searched IP into searched IP database.
	
	>>> addDatabase(IP, CountryName)
	
	Input:
		- IP			= integer number as returned by IP2num
		- CountryName	= full country name
	"""
	global __DB__
	__DB__[IP] = CountryName
	# set a timer to save DB to file?

###################
def IP2num(IP):
	"""
	Converts a string IP (xx.xx.xx.xx) into an integer number.
	
	>>> IPnum = IP2num(IP)
	
	Input:
		- IP		= IP string
	Output:
		- IPnum		= IP equivalent integer number
	"""
	if type(IP) is StringType:
		IPnum = map(int, IP.split('.'))
		return 16777216*IPnum[0] + 65536*IPnum[1] + 256*IPnum[2] + IPnum[3]
	elif type(IP) is IntType:
		return IP
	else:
		raise TypeError, "! IP should be either a string or a long integer"

###################
def num2IP(IPnum):
	"""
	Converts a string IP (xx.xx.xx.xx) into an integer number.
	
	>>> IP = num2IP(IPnum)
	
	Input:
		- IPnum		= IP equivalent integer number
	Output:
		- IP		= IP string
	"""
	if type(IPnum) is not IntType:
		raise TypeError, "! IP should be a long integer"
	IP = ''
	p = 16777216
	for k in range(4):
		Value = int(IPnum / p)
		IPnum -= Value * p
		p /= 256
		if k > 0:
			IP += "."
		IP += str(Value)
	return IP

###################
def guessIPcountry(IP, noInternet=False, recompute='none'):
	"""
	Guesses the IP country using local database and eventually an
	Internet whois service. Local database is always queried before
	any Internet service. Internet services are queried if local
	database cannot answer the query.
	
	>>> CountryName = guessIPcountry(IP, noInternet=False, recompute='none')
	
	Input:
		- IP			= IP string
		- noInternet	= never use Internet whois service
		- recompute		= 'none', 'all' or 'unknown'
	Output:
		- CountryName	= full country name, '?' if not found in databases
	
	Meaning of the recompute values:
		- 'none'	= never search IP in whois databases if the IP
					  was previously searched (searched IP database)
		- 'all'		= always search IP in whois databases
		- 'unknown'	= as 'none' but IP that was previously unknown
					  in databases (CountryName = '?') are searched
					  again
	"""
	global FromIPList, ToIPList, IPCountryList, IPCountryTable, WhoisServer, __DB__
	
	# check args
	if recompute not in ('none', 'all', 'unknown'):
		raise ValueError, "recompute argument should be 'none', 'all' or 'unknown'"
	
	# search local database
	IPnum = IP2num(IP)
	if recompute != 'all':
		if IPnum in __DB__:
			if recompute == 'none' or __DB__[IPnum] != '?':
				return __DB__[IPnum]
	Match = pylab.find(pylab.logical_and(FromIPList < IPnum, ToIPList > IPnum))
	if len(Match):
		CountryName = IPCountryList[Match[0]]
		addDatabase(IPnum, CountryName)
		return CountryName
	
	# try an Internet whois
	CountryName = '?'
	if not noInternet:
		Whois = WhoisConsumer(IP)
		WhoisRequest(Whois, WhoisServer)
		asyncore.loop()
		CountryCode = re.findall('[Cc]ountry:\s+(\w+)', Whois.text)
		if len(CountryCode):
			CountryCode = CountryCode[0]
		else:
			CountryCode = ''
		if CountryCode in IPCountryTable:
			CountryName = IPCountryTable[CountryCode]
	addDatabase(IPnum, CountryName)
	return CountryName

###################
###################
###################

readIPCountryFile()
loadDatabase()
